#!/usr/bin/env python

"""
RSVP instruction classes.

Copyright (C) 2007, 2008 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from micropython.common import Label
from micropython.data import Attr, Const, Instance

def raw(code):
    new_code = []
    for item in code:
        if isinstance(item, Attr):
            new_code.append((item.context and item.context.location, item.value and item.value.location))
        else:
            new_code.append(item)
    return new_code

def name(attr):
    if isinstance(attr, Attr):
        return attr.name
    else:
        return attr

class Instruction:

    "A generic instruction."

    def __init__(self, attr=None):
        self.attr = attr
        self.input = None
        self.source = None # for storage instructions

    def copy(self):
        return self.__class__(self.attr)

    def __repr__(self):
        if self.attr is not None:
            return "%s(%r)%s" % (self.__class__.__name__, self.attr, self.show_input())
        else:
            return "%s()%s" % (self.__class__.__name__, self.show_input())

    def show_input(self):
        if self.input is not None:
            if self.source is not None:
                return " <- (%r, %r)" % (self.input, self.source)
            else:
                return " <- %r" % self.input
        else:
            return ""

class FrameRelativeInstruction(Instruction):

    "An instruction operating on the current frame."

    def __repr__(self):
        return "%s(%r)%s" % (self.__class__.__name__, self.get_operand(), self.show_input())

    def get_operand(self):
        return self.attr.position

FR = FrameRelativeInstruction

class AddressRelativeInstruction(Instruction):

    "An instruction accessing an object's attribute."

    def __repr__(self):
        position = self.get_operand()
        if position is not None:
            return "%s(%r)%s # %s" % (self.__class__.__name__, position, self.show_input(), name(self.attr))
        else:
            return "%s(%r)%s" % (self.__class__.__name__, name(self.attr), self.show_input())

    def get_operand(self):
        return self.attr.position

AR = AddressRelativeInstruction

class AddressInstruction(Instruction):

    "An instruction loading an address directly."

    def __repr__(self):
        location, position, result = self.get_operands()
        if location is not None:
            return "%s(%r)%s # %r, %r (%s)" % (
                self.__class__.__name__, result, self.show_input(), location, position, name(self.attr))
        elif result is not None:
            return "%s(%r)%s # %s" % (
                self.__class__.__name__, result, self.show_input(), name(self.attr))
        else:
            return "%s(...)%s # %s" % (
                self.__class__.__name__, self.show_input(), name(self.attr))

    def get_operands(self):
        if isinstance(self.attr, Attr):
            position = self.attr.position
            location = self.attr.parent.location

            # NOTE: Unpositioned attributes are handled here.

            if location is not None and position is not None:
                result = location + position + 1
            else:
                location = self.attr.parent.name
                position = self.attr.name
                result = None
            return location, position, result
        elif isinstance(self.attr, Label):
            return None, None, self.attr.location
        else:
            return None, None, self.attr.location

    def get_operand(self):
        return self.get_operands()[-1]

Address = AddressInstruction

class ImmediateInstruction(Instruction):

    "An instruction employing a constant."

    def __repr__(self):
        return "%s(%r)%s" % (self.__class__.__name__, self.attr, self.show_input())

    def get_operand(self):
        return self.attr

Immediate = ImmediateInstruction

# Access to stored constant data.

class LoadConst(Address):           "Load the constant, class, function, module from the specified location."

# Access within an invocation frame.

class LoadName(FR):                 "Load the object from the given local attribute/variable."
class StoreName(FR):                "Store the object in the given local attribute/variable."
class LoadTemp(Immediate):          "Load the object from the given temporary location."
class StoreTemp(Immediate):         "Store the object in the given temporary location."

# Access to static data.

class LoadAddress(Address):         "Load the object from the given fixed attribute address."
class StoreAddress(Address):        "Store an object in the given fixed attribute address."
class LoadAddressContext(Address):  "Load the object from the given fixed attribute address, changing the context."
class StoreAddressContext(Address): "Store an object in the given fixed attribute address, changing the context."
class MakeObject(Instruction):      "Make a new object. There isn't a complementary DropObject."

# Access to address-relative data.

class LoadAttr(AR):                 "Load the object from the given attribute."
class StoreAttr(AR):                "Store an object in the given attribute."
class LoadAttrIndex(Immediate):     "Load the object for the attribute with the given index."
class StoreAttrIndex(Immediate):    "Store an object in the attribute with the given index."

# Access to invocation frames in preparation.

class MakeFrame(Instruction):       "Make a new invocation frame."
class DropFrame(Instruction):       "Drop an invocation frame."
class StoreFrame(Immediate):        "Store an argument for the parameter with the given position."
class StoreFrameIndex(Immediate):   "Store an argument for the parameter with the given index."
class LoadCallable(Instruction):    "Load the target of an invocation."
class LoadContext(Instruction):     "Load the context of an invocation."
class CheckFrame(Instruction):      "Check the invocation frame and context for the target."
class CheckSelf(Instruction):       "Check the first argument of an invocation against the target."

# Invocation-related instructions, using a special result "register".

class JumpWithFrame(Instruction):   "Jump, adopting the invocation frame, to the callable found as the current value."
class Return(Instruction):          "Return a value from a subprogram."
class LoadResult(Instruction):      "Load a returned value."

# Branch-related instructions.

class Jump(Address):                "Jump unconditionally."
class JumpIfFalse(Address):         "Jump if the last evaluation gave a false result."
class JumpIfTrue(Address):          "Jump if the last evaluation gave a true result."

# Exception-related instructions, using a special exception "register".

class LoadException(Instruction):   "Load the raised exception."
class RaiseException(Instruction):  "Raise an exception."
class CheckException(Instruction):  "Check the raised exception against another."
class PushHandler(Address):         "Push an exception handler onto the handler stack."
class PopHandler(Instruction):      "Pop an exception handler from the handler stack."

# General instructions.

class TestBoolean(Instruction):     "Test whether the current value is a true value."
class TestIdentityAddress(Address): "Test whether the current value is identical to the given address."

# vim: tabstop=4 expandtab shiftwidth=4
